### rel tools module.
### (mostly rel example code)

import rel
try:
    from subprocess import getoutput # py3
except:
    from commands import getoutput # py2

### helper functions

def notice(*lines):
    print("")
    print("\n".join(lines))

def exit():
    notice("goodbye")
    if rel.running:
        rel.abort()
    else:
        import sys
        sys.exit()

def error(msg, *lines):
    notice("fatal exception!", "error: %s"%(msg,), *lines)
    exit()

### the tools themselves

# rtimer

RT_MP3 = '/var/local/rtimer_elapsed.mp3'
RT_USAGE = 'rtimer [seconds] [minutes] [hours] [update_increment]\n--\nall arguments default to zero. so, rtimer 15 30 2 60 means run for 15 seconds, 30 minutes, and 2 hours, printing a notice every 60 seconds. rtimer 0 5 means run for 5 minutes, printing no incremental updates. when the time runs out, a sound will play on two conditions: there is a readable file at the specified path (default: %s), and mplayer is installed.'%(RT_MP3,)

class Timer(object):
    def __init__(self, s, m, h, interval=0, mp3=RT_MP3):
        try:
            s, m, h, self.interval = int(s), int(m), int(h), int(interval)
        except:
            error("invalid input", "seconds, minutes, hours, and interval must all be integers")
        self.count = 0
        self.goal = s + m * 60 + h * 360
        if self.count == self.goal:
            notice("USAGE: %s"%(RT_USAGE,))
            exit()
        self.mp3 = mp3
        problem = "no sound file path specified"
        if self.mp3:
            import os
            if "command not found" in getoutput("mplayer"):
                self.mp3 = None
                problem = "could not find mplayer!"
            elif not os.path.isfile(mp3):
                self.mp3 = None
                problem = "could not access sound file at %s -- no such file"%(mp3,)
            else:
                try:
                    f = open(mp3)
                    f.close()
                except:
                    self.mp3 = None
                    problem = "could not access sound file at %s -- permission denied"%(mp3,)
        if not self.mp3:
            notice("sound disabled", problem)

    def start(self):
        self.count = 0
        notice("starting countdown to %s"%(self.goal,))
        rel.timeout(1, self.update)
        rel.signal(2, self.stop)
        rel.dispatch()

    def stop(self):
        notice("stopping timer at %s"%(self.count,))
        exit()

    def update(self):
        self.count += 1
        if self.interval and not self.count % self.interval:
            notice("count: %s. goal: %s. completed: %s%%."%(self.count, self.goal, str(self.count/float(self.goal))[:5]))
        if self.count == self.goal:
            self.alarm()
        return True

    def alarm(self):
        notice("time's up!")
        if self.mp3:
            getoutput("mplayer %s"%(self.mp3,))
        self.stop()

### functions for interpreting command-line instructions

# rtimer

def timerCLI():
    from optparse import OptionParser
    parser = OptionParser(RT_USAGE)
    parser.add_option("-m", "--mp3_file_path", dest="mp3", default=RT_MP3, help="location of alarm sound mp3. default: %s"%(RT_MP3,))
    parser.add_option("-v", "--verbose", dest="verbose", default=False, action="store_true", help="run timer in verbose mode")
    options, arguments = parser.parse_args()
    if options.verbose:
        rel.initialize(options=["verbose"])
    try:
        arguments = [int(arg) for arg in arguments]
    except:
        error("non-integer argument", "USAGE: %s"%(RT_USAGE,))
    while len(arguments) < 4:
        arguments.append(0)
    arguments.append(options.mp3)
    Timer(*arguments).start()



import argparse
import dataclasses
import json
import logging
import time
from getpass import getpass
from pprint import pprint
from typing import Callable, Dict, List

import requests
import websocket

from homelypy.devices import Location, SingleLocation, create_device_from_rest_response

WEB_SOCKET_URL = "ws://sdk.iotiliti.cloud"

SDK_URL = "https://sdk.iotiliti.cloud"
AUTHENTICATION_ENDPOINT = "/homely/oauth/token"
LOCATIONS_ENDPOINT = "/homely/locations"
SINGLE_LOCATION_ENDPOINT = "/homely/home"
REFRESH_TOKEN_ENDPOINT = "/homely/oauth/refresh-token"

logger = logging.getLogger(__name__)


class ConnectionFailedException(Exception):
    pass


class AuthenticationFailedException(Exception):
    pass


class Homely:
    def __init__(self, username: str, password: str):
        self.refresh_expires_in = 0
        self.expires_in = 0
        self.authentication_time = 0
        self.access_token = None
        self.refresh_token = None
        self.username = username
        self.password = password

    @staticmethod
    def url(endpoint: str) -> str:
        return SDK_URL + endpoint

    def authenticate(self):
        response = requests.post(
            self.url(AUTHENTICATION_ENDPOINT),
            data={"username": self.username, "password": self.password},
        )
        if response.status_code == 401:
            raise AuthenticationFailedException(response.text)
        if response.status_code != 201:
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def store_authentication_information(self, data: Dict):
        self.access_token = data["access_token"]
        self.authentication_time = time.time()
        self.expires_in = data["expires_in"]
        self.refresh_token = data["refresh_token"]
        self.refresh_expires_in = data["refresh_expires_in"]

    def reauthenticate(self):
        response = requests.post(
            self.url(REFRESH_TOKEN_ENDPOINT),
            data={"refresh_token": self.refresh_token},
        )
        if response.status_code != 201:
            self.refresh_expires_in = -1
            self.expires_in = -1
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def authenticate_if_required(self):
        if time.time() - self.authentication_time > self.refresh_expires_in - 2:
            self.authenticate()
        elif time.time() - self.authentication_time > self.expires_in - 2:
            self.reauthenticate()

    @property
    def authorisation_header(self) -> Dict:
        self.authenticate_if_required()
        return {"Authorization": f"Bearer {self.access_token}"}

    def get_locations(self) -> List[Location]:
        response = requests.get(self.url(LOCATIONS_ENDPOINT), headers=self.authorisation_header)
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        return [
            Location(
                location_data["name"],
                location_data["role"],
                location_data["userId"],
                location_data["locationId"],
                location_data["gatewayserial"],
            )
            for location_data in response.json()
        ]

    def get_location_json(self, location_id) -> dict:
        response = requests.get(
            self.url(SINGLE_LOCATION_ENDPOINT) + f"/{location_id}",
            headers=self.authorisation_header,
        )
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        return response.json()

    def get_location(self, location_id) -> SingleLocation:
        data = self.get_location_json(location_id)
        devices = []
        for device in data["devices"]:
            devices.append(create_device_from_rest_response(device))
        return SingleLocation(
            data["locationId"],
            data["gatewayserial"],
            data["name"],
            data["alarmState"],
            data["userRoleAtLocation"],
            devices,
        )

    def get_web_socket(self, location_id: str, on_message: Callable) -> websocket.WebSocketApp:
        self.authenticate_if_required()
        websocket.enableTrace(True)
        url = f"{WEB_SOCKET_URL}?locationId={location_id}"
        logger.debug(f"Connecting to web socket {url}")
        ws = websocket.WebSocketApp(
            url,
            header={**self.authorisation_header, "locationId": location_id},
            on_open=on_open,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
        )
        return ws


def on_error(ws, error):
    logger.debug(f"Websocket error: {error}")


def on_close(ws, close_status_code, close_msg):
    logger.debug(f"Websocket closed")


def on_open(ws):
    logger.debug(f"Websocket connected")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(prog="Homelypy", description="Query the Homely rest API")
    parser.add_argument("username", help="Same username as in the Homely app")
    args = parser.parse_args()
    password = getpass()

    homely = Homely(args.username, password)
    locations = homely.get_locations()
    for location in locations:
        print(f"Received location '{location}'")
    for location in locations:
        print(f"Full dump for location {location}")
        location_dictionary = homely.get_location_json(location.location_id)
        pprint(location_dictionary)
        print("----------------------------------")
        with open(f"location_{location.name}.json", "w") as o:
            json.dump(location_dictionary, o)
    # location = homely.get_location(locations[0].location_id)
    # logger.debug(f"Received single location '{location}'")
    # for device in location.devices:
    #     logger.debug(f"Device: {device} is of type {device.__class__}")
    #
    # location = homely.get_location(locations[0].location_id)
    # ws = homely.get_web_socket(location.location_id, lambda data: logger.debug(f"Received data: {data}"))
    # import rel
    #
    # ws.run_forever(
    #     dispatcher=rel, reconnect=5
    # )  # Set dispatcher to automatic reconnection, 5 second reconnect delay if connection closed unexpectedly
    # rel.signal(2, rel.abort)  # Keyboard Interrupt
    # rel.dispatch()



import time
import signal

EV_PERSIST = 16
EV_READ = 2
EV_SIGNAL = 8
EV_TIMEOUT = 1
EV_WRITE = 4
noadd = "this is a do-not-add order from your mother, an Event object"

def contains(mode,bit):
    return mode&bit==bit

class Event(object):
    def __init__(self,registrar,cb,arg,evtype,handle):
        self.registrar = registrar
        self.cb = cb
        self.arg = arg
        self.timeout = self.registrar.timeout(None,self.callback)
        self.evtype = evtype or 1
        self.handle = handle
        self.children = []
        self.spawn_children()

    def spawn_children(self):
        persist = contains(self.evtype,EV_PERSIST)
        if contains(self.evtype,EV_SIGNAL):
            self.children.append(self.registrar.signal(self.handle,self.callback,noadd))
        if contains(self.evtype,EV_READ):
            tmp = self.registrar.read(self.handle,self.callback,noadd)
            if persist:
                tmp.persistent()
            self.children.append(tmp)
        if contains(self.evtype,EV_WRITE):
            tmp = self.registrar.write(self.handle,self.callback,noadd)
            if persist:
                tmp.persistent()
            self.children.append(tmp)

    def add(self, delay=None):
        if delay is not None:
            self.timeout.add(delay)
        for child in self.children:
            child.add()

    def delete(self):
        self.timeout.delete()
        for child in self.children:
            child.delete()

    def pending(self):
        for child in self.children:
            if child.pending():
                return 1
        return self.timeout.pending()

    def callback(self):
        self.cb(self,self.handle,self.evtype,self.arg)

class SocketIO(object):
    def __init__(self, registrar, evtype, sock, cb, *args):
        self.registrar = registrar
        self.evtype = evtype
        self.sock = sock
        self.fd = sock
        if hasattr(self.fd,'fileno'):
            self.fd = self.fd.fileno()
        self.cb = cb
        self.args = args
        self.persist = False
        self.active = 0
        if noadd in self.args:
            self.args = ()
            return
        self.timeout = self.registrar.timeout(None,self.callback)
        self.add()

    def __repr__(self):
        cbname = self.cb.__name__
        if hasattr(self.cb,"im_class"):
            cbname = self.cb.__self__.__class__.__name__ + "." + cbname
        return '<SocketIO Object | Callback:"%s">'%cbname

    def persistent(self):
        self.persist = True

    def add(self, delay=None):
        if delay is not None:
            self.timeout.add(delay)
        self.registrar.add(self)
        self.active = 1

    def delete(self):
        self.registrar.remove(self)
        self.active = 0

    def dereference(self):
        if self.pending():
            self.delete()
        self.cb = None
        self.args = None
        self.timeout.delete(True)

    def pending(self):
        return self.active

    def callback(self):
        if not self.cb(*self.args) and not self.persist and self.active:
            self.delete()

class Signal(object):
    def __init__(self, registrar, sig, cb, *args):
        self.registrar = registrar
        self.sig = sig
        self.default = signal.getsignal(self.sig)
        self.cb = cb
        self.args = args
        if noadd in self.args:
            self.args = ()
            return
        self.timeout = self.registrar.timeout(None,self.callback)
        self.add()

    def __repr__(self):
        cbname = self.cb.__name__
        if hasattr(self.cb,"im_class"):
            cbname = self.cb.__self__.__class__.__name__ + "." + cbname
        return '<Signal Object | Callback:"%s">'%cbname

    def add(self, delay=None):
        if delay is not None:
            self.timeout.add(delay)
        signal.signal(self.sig,self.callback)
        self.active = 1
        self.registrar.signal_add(self)

    def delete(self):
        self.reset()
        self.active = 0
        self.registrar.signal_remove(self.sig)

    def reset(self):
        signal.signal(self.sig,self.default)

    def pending(self):
        return self.active

    def callback(self,*args):
        self.cb(*self.args)
        self.registrar.error_check = True

class Timer(object):
    def __init__(self, registrar, delay, cb, *args):
        self.registrar = registrar
        self.cb = cb
        self.args = args
        if noadd in self.args:
            self.args = ()
            return
        self.add(delay)

    def __repr__(self):
        cbname = self.cb.__name__
        if hasattr(self.cb,"im_class"):
            cbname = self.cb.__self__.__class__.__name__ + "." + cbname
        return '<Timer Object | Callback:"%s">'%cbname

    def add(self, delay=None):
        self.delay = delay
        self.expiration = None
        if self.delay is not None:
            self.expiration = time.time()+self.delay
            self.registrar.add_timer(self)

    def delete(self, dereference=False):
        self.expiration = None
        self.registrar.remove_timer(self)
        if dereference:
            self.cb = None
            self.args = None

    def pending(self):
        if self.expiration:
            return 1
        return 0

    def check(self, t=None):
        if not self.pending():
            return False
        if (t or time.time()) >= self.expiration:
            if self.cb(*self.args):
                self.add(self.delay)
                return True
            return False
        return True